// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JFPPR.
// 
// JFPPR is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JFPPR is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JFPPR.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.graph.fppr;

import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
 * This is a structure for representing a graph capable of performing Fully
 * Personalized PageRank efficiently through random walk sampling.
 * <p>This structure can be used as follows:
 * <ol><li>A Graph instance is instantiated with the number of {@link Vertex}
 * elements in the graph.</li><li>Following this, all outgoing edges are
 * specified with {@link Vertex#addOutgoingEdge(Vertex, float)}.</li><li>After
 * the edges are set, random walks are simulated on the graph using the
 * {@link #walk(int)} method.</li><li>Finally, for Vertex elements for which
 * a Personalized PageRank is desired, the {@link #rank(int,int,List)}
 * is used.</li>
 * </ol>
 */
public final class Graph {
    private Random r = new Random();
    private Vertex[] vertices, sorted_vertices;

    /**
     * The probability of ending a random walk segment and either resetting
     * to a source vertex or jumping to another random vertex.
     */
    public static final float RESET_PROBABILITY = 0.2f;

    /**
     * @param vertex_count number of {@link Vertex} elements in the graph
     */
    public Graph(int vertex_count) {
        int i;
        vertices = new Vertex[vertex_count];
        sorted_vertices = new Vertex[vertex_count];
        for(i = 0;i < vertex_count;i++)
            sorted_vertices[i] = vertices[i] = new Vertex(i);
    }

    /**
     * @param vertex_id the id of the vertex to retrieve.  Starting at zero
     * and going to the number of {@link Vertex} elements specified in the
     * constructor (exclusive).
     * @return the {@link Vertex} element corresponding to the specified vertex
     * id.
     */
    public Vertex vertex(int vertex_id) {
        return vertices[vertex_id];
    }

    /**
     * @return the number of {@link Vertex} elements specified in the
     * constructor.
     */
    public int vertices() {
        return vertices.length;
    }

    /**
     * Simulate random walks.
     *
     * @param walks_per_vertex the number of random walks per Vertex to take.
     * This number should be proportional to the logarithm of the number of
     * vertices.  For graphs that will fit in main memory, this number will
     * probably not need exceed 10 or 15.  A walk is a successive number of
     * weighted random transitions following outgoing edges until a reset
     * is encountered.
     *
     * @see #RESET_PROBABILITY
     */
    public void walk(int walks_per_vertex) {
        int i, I = vertices.length, w;
        for(w = 0;w < walks_per_vertex;w++)
            for(i = 0;i < I;i++)
                walk(vertices[i]);
    }

    private void walk(Vertex start) {
        Step last_step = new Step(start);
        while(r.nextFloat() > RESET_PROBABILITY)
            last_step = last_step.next_step = new Step(last_step.vertex.sampleNeighbor());
    }

    /**
     * Obtain a descending ranking of nearest neighbors according to a
     * Personalized PageRank of this Graph.
     *
     * @param start_vertex_id the id of the source {@link Vertex} for which to
     * perform the Personalized PageRank and obtain neighbors.  This number goes
     * from zero to the number of vertices specified in the constructor
     * (exclusive)
     * @param walk_length the total number of walk steps taken.  Resets to the
     * {@link Vertex} specified by the start_vertex_id are also counted as a
     * walk step.  A reasonable number for this is on the order of 10,000.
     * @param result a user-specified destination for storing the ranked
     * {@link Vertex} ids such that the nearest neighbor will be stored first.
     *
     * @return the user-specified destination for storing the ranked
     * {@link Vertex} ids (returned for convenience).
     */
    public List<Integer> rank(int start_vertex_id, int walk_length, List<Integer> result) {
        int i, I = vertices.length, steps = 1;
        Step step;
        Vertex last_vertex = vertices[start_vertex_id];
        result.clear();
        for(i = 0;i < I;i++)
            vertices[i].reset();
        last_vertex.visit();
        while(steps++ < walk_length)
            if(r.nextFloat() <= RESET_PROBABILITY)
                (last_vertex = vertices[start_vertex_id]).visit();
            else {
                if((step = last_vertex.nextSegmentStart()) != null) {
                    while((step = step.next_step) != null)
                        step.vertex.visit();
                    (last_vertex = vertices[start_vertex_id]).visit();
                } else
                    (last_vertex = last_vertex.sampleNeighbor()).visit();
            }
        Arrays.sort(sorted_vertices);
        for(Vertex v : sorted_vertices)
            if(v.visits() == 0)
                break;
            else if(v.id != start_vertex_id)
                result.add(v.id);
        return result;
    }

    protected static class Step {
        Vertex vertex;
        Step next_step = null;

        Step(Vertex v) {
            vertex = v;
            v.addSegment(this);
        }
    }
}
