// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JFPPR. 
// 
// JFPPR is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JFPPR is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JFPPR.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.graph.fppr;

import com.pawjaw.graph.fppr.Graph.Step;
import java.util.Random;

/**
 * Each element in the PageRank {@link Graph} is represented by an instance of
 * this class.  Vertex elements are by the Graph instance when the Graph is
 * instantiated.  The number of Vertex elements is specified when the Graph is
 * instantiated.  Vertex elements are labeled by contiguous increasing integer ids
 * starting at zero.  The process for adding outgoing edges from a Vertex is by
 * retrieving the source Vertex from the Graph instance, then specifying a
 * destination Vertex also retrieved from the Graph instance as well as a
 * non-negative weight.
 *
 * @see Graph
 */
public final class Vertex implements Comparable<Vertex> {
    /**
     * The id of this vertex.
     */
    public final int id;

    private int visits = 0;
    private float outgoing_edge_weight_sum = 0;
    private Segment next_segment = null;
    private Segment last_segment = null;
    private Segment first_segment = null;
    private OutgoingEdge last_outgoing_edge = null;
    private OutgoingEdge first_outgoing_edge = null;

    private static final Random R = new Random();

    protected Vertex(int id) {
        this.id = id;
    }

    /**
     * Add a directed outgoing edge from this Vertex to a destination Vertex.
     * Use the {@link Graph#vertex} method to retrieve source and destination
     * Vertex elements.
     * <p>Here is an example of creating a Graph with 10 Vertex
     * elements and adding an outgoing edge from Vertex with vertex id: 7 to
     * Vertex with vertex id: 2 with transition weight 0.5:
     * <p><code>
     * Graph g = new Graph(10);<br>
     * Vertex source = g.vertex(7);<br>
     * Vertex destination = g.vertex(2);<br>
     * source.addOutgoingEdge(destination, 0.5);
     * </code>
     *
     * @param destination the destination of this outgoing edge
     * @param weight a non-negative transition weight for sampling walk steps
     *
     * @see Graph#vertex(int)
     */
    public void addOutgoingEdge(Vertex destination, float weight) {
        outgoing_edge_weight_sum += weight;
        if(first_outgoing_edge == null)
            first_outgoing_edge = last_outgoing_edge = new OutgoingEdge(destination, weight);
        else
            last_outgoing_edge = last_outgoing_edge.next_outgoing_edge = new OutgoingEdge(destination, weight);
    }

    protected void addSegment(Step step) {
        if(first_segment == null)
            next_segment = first_segment = last_segment = new Segment(step);
        else
            last_segment = last_segment.next_segment = new Segment(step);
    }

    protected void reset() {
        visits = 0;
        next_segment = first_segment;
    }

    protected void visit() {
        visits++;
    }

    protected int visits() {
        return visits;
    }

    protected Step nextSegmentStart() {
        Segment _next_segment;
        if(next_segment == null)
            return null;
        _next_segment = next_segment;
        next_segment = next_segment.next_segment;
        return _next_segment.segment_start;
    }

    protected Vertex sampleNeighbor() {
        float r = R.nextFloat() * outgoing_edge_weight_sum;
        OutgoingEdge next_outgoing_edge = first_outgoing_edge;
        while(next_outgoing_edge != null)
            if((r -= next_outgoing_edge.weight) <= 0)
                return next_outgoing_edge.destination;
            else
                next_outgoing_edge = next_outgoing_edge.next_outgoing_edge;
        // if no neighbors, treat as if exists single loop edge back
        return this;
    }

    public int compareTo(Vertex o) {
        return visits < o.visits ? 1 : -1;
    }

    private static class OutgoingEdge {
        float weight;
        Vertex destination;
        OutgoingEdge next_outgoing_edge = null;

        OutgoingEdge(Vertex destination, float weight) {
            this.destination = destination;
            this.weight = Math.max(0, weight);
        }
    }

    private static class Segment {
        Step segment_start;
        Segment next_segment = null;

        Segment(Step step) {
            this.segment_start = step;
        }
    }
}
